#define WIDTH	320
#define HEIGHT	240

#define BASE	0x00200000

	.text

	.align	2
start:

	.word	0x464c457f			// e_ident
dev_mem_filename:
	.ascii "/dev/mem"
	.word	0x00000000
	.word	0x00280002			// e_machine | e_type
ball_bat_color_x4:
	.word	0x01010101			//(e_version)
	.word	entry-start+BASE	// e_entry
	.word	prog_header-start	// e_phoff
prog_header:
	.word	0x00000001			//(e_shoff)						// p_type
	.word	0x00000000			//(e_flags)						// p_offset
	.word	BASE				// e_phentsize |(e_ehsize)		// p_vaddr
	.word	0x00000001			//(e_shentsize)| e_phnum		//(p_paddr)
lcrng_params:
	.word	end-start			//(e_shstrndx) |(e_shnum)		// p_filesz			//(lcrng_seed)
	.word	0x0019660d											//(p_memsz)			// lcrng_A
	.word	0x3c6ef35f											// p_flags (1<<1)	// lcrng_B

reglist:
	.word	0x28da02ab		// store 0x02ab in gpio 0x28da		// p_flags - bit 2 must be set
	.word	0x290c0140		// store 320    in gpio 0x290c		//(p_align)
	.word	0x29580000		// set palette starting at entry 0
	.word	0x295affff
	.word	0x295a0000

mmap_params:
	.word	0x00000000		// addr (ignored)
	.word	0x1e000			// size
	.word	3				// prot = PROT_READ | PROT_WRITE
	.word	1				// flags = MAP_SHARED
mmap_params_fd:
	.word	3				// fd
mmap_params_offset:
	.word	0x03101000		// offset


	.align	2
	.global	entry
entry:
	// open /dev/mem
	adr		r0, dev_mem_filename
	mov		r1, #2				// O_RDWR
	swi		#0x900005			// sys_open

	str		r0, mmap_params_fd

	// mmap the vram
	adr		r0, mmap_params
	swi		#0x90005a			// old_mmap
	mov		r11, r0				// r11 = vram addr

	// mmap the gpio regs
	adr		r0, mmap_params
	mov		r2, #0xc0000000
	str		r2, mmap_params_offset
	swi		#0x90005a			// old_mmap
	add		r12, r0, #0x1180	// r12 = gpio addr + 0x1180


	// set gpio regs
	adr		r1, reglist
1:
	ldr		r2, [r1], #4
	movs	r7, r2, lsr #16
	strneh	r2, [r0, r7]		// r0 is still the gpio base addr, for now
	strneh	r2, [r0, r7]		// set everything twice - reduces number of entries for palette writing (!)
	bne		1b


	// allocate working buffers in cached memory (on the stack)
	sub		r10, sp, #2*WIDTH*HEIGHT
	mov		r9, #WIDTH*HEIGHT			// offset from current buffer to next buffer


init:

	// init the rng
	adr		r5, lcrng_params
	ldmia	r5, { r6, r7, r8 }


	// fill buffer randomly
	mov		r0, #WIDTH*HEIGHT
	ldr		r3, ball_bat_color_x4
1:
	// crank the rng
	mlas	r6, r7, r6, r8
	
	// massage the result to get a four element word
//	and		r2, r6, #0x0f000000			// 0000abcd 00000000 00000000 00000000
//	orr		r2, r2, r2, ror #9			// 0000abcd 00000abc d0000000 00000000
//	orr		r2, r2, r2, ror #18			// bcd0abcd 00000abc d00000ab cd00000a
//	and		r2, r2, r3
	eor		r2, r6, r6, ror #15
	and		r2, r3, r2, ror #7

	subs	r0, r0, #4
	str		r2, [r10, r0]	// randomize initial configuration
	bne		1b


	// store random seed
	stmia	r5, { r6, r7, r8 }


main_loop:
	
	// read input - we need to do this before render so we can effectively double-vsync
	ldrh	r8, [r12, #4]		// read register C
	mvn		r8, r8
	ands	r7, r8, #0x1000		// A = slow
slow:
	blne	vsync
	movnes	r7, r7, lsr #1
	bne		slow

	// render
	mov		r0, #WIDTH*HEIGHT
1:
	subs	r0, r0, #4
	ldr		r1, [r10, r0]
	str		r1, [r11, r0]
	bne		1b

	// more input processing - r8 is still valid
	tst		r8, #0x4000			// X = pause
	bne		main_loop
	tst		r8, #0x2000			// B = reset
	bne		init
	tst		r8, #0x0100			// start = quit
	swine	#0x900001

	// update next generation buffer
	mov		r8, #HEIGHT*WIDTH
	sub		r8, r8, #WIDTH*2
	add		r1, r10, #WIDTH		// note that r1 points at pixel to the left of current - we can't add #WIDTH+1 :(
1:
	ldrb	r5, [r1], #-WIDTH
	ldrb	r6, [r1], #WIDTH*2
	add		r5, r5, r6
	ldrb	r6, [r1], #2
	add		r5, r5, r6
	ldrb	r6, [r1], #-WIDTH
	add		r5, r5, r6
	ldrb	r6, [r1], #-WIDTH
	add		r5, r5, r6
	ldrb	r6, [r1], #-1
	add		r5, r5, r6
	ldrb	r6, [r1], #WIDTH*2
	add		r5, r5, r6
	ldrb	r6, [r1], #-WIDTH
	add		r5, r5, r6
	subs	r5, r5, #3
	ldrneb	r7, [r1]				// r1 is now pointing at the current pixel
	addnes	r5, r5, r7
	movne	r5, #0
	moveq	r5, #1
	strb	r5, [r1, r9]			// write pixel in next buffer

	subs	r8, r8, #1
	bne		1b


	// flip buffers
	add		r10, r10, r9
	rsb		r9, r9, #0


	// sync vertical edges
	mov		r0, #HEIGHT
	add		r1, r10, #WIDTH/2
1:
	subs	r0, r0, #1
	ldrb	r2, [r1, #(WIDTH/2)-2]
	strb	r2, [r1, #-WIDTH/2]
	ldrb	r2, [r1, #-(WIDTH/2-1)]
	strb	r2, [r1, #(WIDTH/2)-1]
	add		r1, r1, #WIDTH
	bne		1b

	// sync horizontal edges
	mov		r0, #WIDTH
	mov		r1, r10
	add		r2, r1, #WIDTH*HEIGHT
1:
	subs	r0, r0, #4
	ldr		r3, [r2, #-WIDTH*2]
	str		r3, [r1], #4
	ldr		r3, [r1, #WIDTH-4]
	str		r3, [r2, #-WIDTH]
	add		r2, r2, #4
	bne		1b

	adr		lr, main_loop


vsync:
	adr		r2, vsync_loop		// start off looping to vsync_loop
vsync_loop:
	ldr		r1, [r12, #2]
	tst		r1, #0x10
	movne	pc, r2				// when we leave vblank, stop looping to r2...
	mov		r2, lr				// ... and set r2 to return address...
	b		vsync_loop			// ... and loop to vsync_loop explicitly.
								// Now when we reenter vblank, the movne will fire
								// once again, and we'll have returned.

end:	// mark end of program for header calculations

